#include <stdio.h>
#include <CoreServices/CoreServices.h>
#include <CoreAudio/CoreAudio.h>
#include <AudioUnit/AudioUnit.h>
#include <AudioToolbox/AudioToolbox.h>
#include <unistd.h>
#include <pthread.h>
#include "driver.h"

#define SOFTBUFSIZE 98304

OSStatus MyFileRenderProc(void *inRefCon, AudioUnitRenderActionFlags    *inActionFlags,
	const AudioTimeStamp *inTimeStamp, UInt32 inBusNumber,
	UInt32 inNumFrames, AudioBufferList *ioData);
OSStatus MyACComplexInputProc   (AudioConverterRef                        inAudioConverter,
	UInt32                               *ioNumberDataPackets,
	AudioBufferList                                   *ioData,
	AudioStreamPacketDescription  **outDataPacketDescription,
	void*                                          inUserData);
int copyBuffer(char *buffer, int bytes);
void convertToAIFF(char *file);
void *receiveKeyInput(void* param);
void setDefaultMuteChannel(int mask);

typedef struct {
	char FORM[4];
	int formChunkSize;
	char AIFF[4];
	char COMM[4];
	int commChunkSize;
	short numChannels;
	unsigned int numSampleFrames;
	short sampleSize;
	int sampleRate1;
	int sampleRate2;
	short sampleRate3;
	char SSND[4];
	int ssndChunkSize;
	unsigned int offset;
	unsigned int blockSize;
} AIFFHeader;

static AudioUnit outputUnit;
static AudioConverterRef converter;
static char *buffer1;
static char *buffer2;
static char *buffer1Top;
void *gSourceBuffer;
static int bufferIsFull;
static int buffer1Rest;
static int buffer2Rest;
static int totalSize;
static int playedSize;
static int totalTime;
static int currentTime;
static int lastBuffer;
static int playDone;
static int first;
static int convertMode;
static FILE *fpw;
static pthread_t tid;
static int channelMask;

int main(int argc, char *argv[])
{
	OSStatus	err ;
	AudioStreamBasicDescription inFormat;
	AudioStreamBasicDescription outFormat;
	AURenderCallbackStruct  renderCallback;
	Boolean outWritable;
	UInt32		count ;
	int i=0;
	PSFINFO *psfinfo;
	FILE *fp;
	char buf[4], opt;
	extern char *optarg ;
	extern int optind;
	channelMask = 16777215;
	
	while((opt = getopt(argc,argv,"o:m:")) != -1) {
		switch(opt) {
		  case 'o' :
			convertMode = 1;
			if(!(fpw=fopen(optarg,"wb"))) {
				fprintf(stderr,"can't open file.\n");
				return 0;
			}
			i++;
			break;
		  case 'm' :
			channelMask = atoi(optarg);
			break;
		}
		i++;
	}
	
	if(argc <= i+1) {
		printf("usage: %s <PSF file>\n",argv[0]);
		return 0;
	}
	
	if(!(fp=fopen(argv[optind],"rb"))) {
		fprintf(stderr,"can't open file.\n");
		return 0;
	}
	if(fread(buf,1,4,fp)!=4) {
		fprintf(stderr,"can't read file.\n");
		fclose(fp);
		return 0;
	}
	fclose(fp);
	if(memcmp(buf,"PSF\x01",4))	// Only allow for PSF1 files for now.
	{
		fprintf(stderr,"file is not PSF1 format.\n");
		return 0;
	}
	
	if(convertMode) {
		convertToAIFF(argv[optind]);
		return 0;
	}
	
	ComponentDescription desc;
	Component comp;
	
	desc.componentType = kAudioUnitType_Output; 
	desc.componentSubType = kAudioUnitSubType_DefaultOutput;
	desc.componentManufacturer = kAudioUnitManufacturer_Apple;
	desc.componentFlags = 0;
	desc.componentFlagsMask = 0;
	comp = FindNextComponent(NULL, &desc);
	if (comp == NULL) {
		fprintf(stderr, "can't find default output unit\n");
		return 0;
	}
	
	err = OpenAComponent(comp, &outputUnit);
	if (err)  {
		fprintf(stderr, "can't open default output unit\n");
		return 0;
	}
	
	err = AudioUnitGetPropertyInfo(outputUnit, kAudioUnitProperty_StreamFormat, kAudioUnitScope_Output, 0, &count, &outWritable);
	err = AudioUnitGetProperty (outputUnit, kAudioUnitProperty_StreamFormat, kAudioUnitScope_Output, 0, &outFormat, &count);
	if(err != noErr) {
		fprintf(stderr, "AudioUnitGetProperty failed.\n");
		return 0;
	}
	AudioUnitSetProperty (outputUnit, kAudioUnitProperty_StreamFormat, kAudioUnitScope_Input, 0, &outFormat, count);
	if(err != noErr) {
		fprintf(stderr, "AudioUnitSetProperty failed.\n");
		return 0;
	}
	
	err = AudioUnitInitialize(outputUnit);
	if(err != noErr) {
		fprintf(stderr, "AudioUnitInitialize failed.\n");
		return 0;
	}
	
	inFormat.mSampleRate = 44100.0;
	inFormat.mBytesPerPacket = 4;
	inFormat.mFramesPerPacket = 1;
	inFormat.mBytesPerFrame = 4;
	inFormat.mChannelsPerFrame = 2;
	inFormat.mBitsPerChannel = 16;
	inFormat.mFormatID = kAudioFormatLinearPCM;
	inFormat.mFormatFlags = kLinearPCMFormatFlagIsSignedInteger | kLinearPCMFormatFlagIsPacked | kLinearPCMFormatFlagIsBigEndian;
	
	err = AudioConverterNew(&inFormat, &outFormat, &converter);
	if(err != noErr) {
		fprintf(stderr, "AudioConverterNew failed.\n");
		return 0;
	}
	
	memset(&renderCallback, 0, sizeof(AURenderCallbackStruct));
	renderCallback.inputProc = MyFileRenderProc;
	renderCallback.inputProcRefCon =0;
	err = AudioUnitSetProperty (outputUnit, kAudioUnitProperty_SetRenderCallback, kAudioUnitScope_Input, 0, &renderCallback, sizeof(AURenderCallbackStruct));
	if(err != noErr) {
		fprintf(stderr, "AudioUnitSetProperty failed.\n");
		return 0;
	}
	
	if((buffer1 = (char *)malloc(SOFTBUFSIZE)) == NULL) {
		fprintf(stderr,"malloc error\n");
		return 0;
	}
	if((buffer2 = (char *)malloc(SOFTBUFSIZE)) == NULL) {
		fprintf(stderr,"malloc error\n");
		return 0;
	}
	
	bufferIsFull = 0;
	buffer1Rest = SOFTBUFSIZE;
	buffer2Rest = 0;
	buffer1Top = buffer1;
	lastBuffer = 0;
	playDone = 0;
	first = 1;
	playedSize = 0;
	currentTime = -1;
	
	psfinfo = sexy_load(argv[optind]);
	if(psfinfo->title) printf("Title: %s\n",psfinfo->title);
	
	totalSize = psfinfo->length / 1000 * 44100 * 4;
	totalTime = psfinfo->length / 1000;
	setDefaultMuteChannel(channelMask);
	
	pthread_create(&tid,NULL,receiveKeyInput,NULL);
	pthread_detach(tid);
	
	sexy_execute();
	
	memset(buffer1Top, 0, buffer1Rest);
	bufferIsFull = 1;
	lastBuffer = 1;
	
	while(!playDone) {
		usleep(10000);
	}
	
	sexy_stop();
	
	AudioConverterDispose(converter);
	AudioOutputUnitStop(outputUnit);
	AudioUnitUninitialize (outputUnit);
	CloseComponent(outputUnit);
	
	return 0;
}

void convertToAIFF(char *file)
{
	PSFINFO *psfinfo;
	AIFFHeader header;
	
	playedSize = 0;
	currentTime = -1;
	psfinfo = sexy_load(file);
	if(psfinfo->title) printf("Title: %s\n",psfinfo->title);
	totalSize = psfinfo->length / 1000 * 44100 * 4;
	
	setDefaultMuteChannel(channelMask);
	
	memcpy(header.FORM, "FORM", 4);
	fwrite(&header.FORM, 1, 4, fpw);
	header.formChunkSize = totalSize + 0x2e;
	fwrite(&header.formChunkSize, 4, 1, fpw);
	memcpy(header.AIFF, "AIFF", 4);
	fwrite(&header.AIFF, 1, 4, fpw);
	memcpy(header.COMM, "COMM", 4);
	fwrite(&header.COMM, 1, 4, fpw);
	header.commChunkSize = 0x12;
	fwrite(&header.commChunkSize, 4, 1, fpw);
	header.numChannels = 2;
	fwrite(&header.numChannels, 2, 1, fpw);
	header.numSampleFrames = totalSize/4;
	fwrite(&header.numSampleFrames, 4, 1, fpw);
	header.sampleSize = 0x10;
	fwrite(&header.sampleSize, 2, 1, fpw);
	header.sampleRate1 = 0x400eac44;
	fwrite(&header.sampleRate1, 4, 1, fpw);
	header.sampleRate2 = 0;
	fwrite(&header.sampleRate2, 4, 1, fpw);
	header.sampleRate3 = 0;
	fwrite(&header.sampleRate3, 2, 1, fpw);
	memcpy(header.SSND, "SSND", 4);
	fwrite(&header.SSND, 1, 4, fpw);
	header.ssndChunkSize = totalSize + 8;
	fwrite(&header.ssndChunkSize, 4, 1, fpw);
	header.offset = 0;
	fwrite(&header.offset, 4, 1, fpw);
	header.blockSize = 0;
	fwrite(&header.blockSize, 4, 1, fpw);
	
	sexy_execute();
	sexy_stop();
	
	header.formChunkSize = playedSize + 0x2e;
	header.numSampleFrames = playedSize/4;
	header.ssndChunkSize = playedSize + 8;
	fseek(fpw, 4, SEEK_SET);
	fwrite(&header.formChunkSize, 4, 1, fpw);
	fseek(fpw, 0xe, SEEK_CUR);
	fwrite(&header.numSampleFrames, 4, 1, fpw);
	fseek(fpw, 0x10, SEEK_CUR);
	fwrite(&header.ssndChunkSize, 4, 1, fpw);
	
	fclose(fpw);
}

void sexyd_update(char *Buffer, int count)
{
	int copied,percent,i;
	if(!convertMode) {
		while(count) {
			while(bufferIsFull) usleep(10000);
			copied = copyBuffer(Buffer,count);
			count = count - copied;
			Buffer = Buffer + copied;
		}
	}
	else {
		fwrite(Buffer,1,count,fpw);
		playedSize = playedSize + count;
		
		percent = (int)(100.0*playedSize/totalSize);
		if(percent > 100) percent = 100;
		if(percent != currentTime) {
			printf("\r|");
			for(i=0;i<20;i++) {
				if(percent/5 > i)
					printf("=");
				else if(percent/5 == i)
					printf(">");
				else printf("-");
			}
			printf("| %d%%",percent);
			fflush(stdout);
			currentTime = percent;
		}
	}
}

int copyBuffer(char *buffer, int bytes)
{
	int tmp;
	if(buffer1Rest > bytes) {
		memcpy(buffer1Top, buffer, bytes);
		buffer1Top = buffer1Top + bytes;
		buffer1Rest = buffer1Rest - bytes;
		return bytes;
	}
	memcpy(buffer1Top, buffer, buffer1Rest);
	buffer1Top = buffer1Top + buffer1Rest;
	tmp = buffer1Rest;
	buffer1Rest = 0;
	bufferIsFull = 1;
	if(first) {
		AudioOutputUnitStart(outputUnit);
		first = 0;
	}
	return tmp;
}


OSStatus MyFileRenderProc(void *inRefCon, AudioUnitRenderActionFlags    *inActionFlags,
                        const AudioTimeStamp *inTimeStamp, UInt32 inBusNumber,
                        UInt32 inNumFrames, AudioBufferList *ioData)
{
	OSStatus err= noErr;
	void *inInputDataProcUserData=NULL;
	AudioStreamPacketDescription* outPacketDescription =NULL;
	
	err = AudioConverterFillComplexBuffer(converter, MyACComplexInputProc ,inInputDataProcUserData , &inNumFrames, ioData, outPacketDescription);
	if(!playDone && err != noErr) {
		fprintf(stderr, "AudioConverterFillComplexBuffer failed.\n");
		return 0;
	}
	
	return err;
}

OSStatus MyACComplexInputProc   (AudioConverterRef                        inAudioConverter,
                                 UInt32                               *ioNumberDataPackets,
                                 AudioBufferList                                   *ioData,
                                 AudioStreamPacketDescription  **outDataPacketDescription,
                                 void*                                          inUserData)
{
	OSStatus    err = noErr;
	UInt32  bytesCopied = 0;
	int i, sec, percent;
	
	ioData->mBuffers[0].mData = NULL;      
	ioData->mBuffers[0].mDataByteSize = 0;
	
	if(!buffer2Rest && lastBuffer && !bufferIsFull) {
		playDone = 1;
		return noErr;
	}
	
	if(!buffer2Rest) {
		while(!bufferIsFull) usleep(10000);
		memcpy(buffer2, buffer1, SOFTBUFSIZE);
		buffer1Rest = SOFTBUFSIZE;
		buffer1Top = buffer1;
		bufferIsFull = 0;
		buffer2Rest = SOFTBUFSIZE;
	}
	
	//printf("packet size: %d\n",*ioNumberDataPackets);
	
	if (*ioNumberDataPackets > buffer2Rest / 4) 
		*ioNumberDataPackets = buffer2Rest / 4;
	
	if (*ioNumberDataPackets)
	{
		if (gSourceBuffer != NULL) {
			free(gSourceBuffer);
			gSourceBuffer = NULL;
		}
		
		bytesCopied = *ioNumberDataPackets * 4;  
		gSourceBuffer = (void *) calloc (1, bytesCopied);             
		memcpy(gSourceBuffer, buffer2 + SOFTBUFSIZE - buffer2Rest, bytesCopied);
		
		buffer2Rest = buffer2Rest - bytesCopied;
		
		ioData->mBuffers[0].mData = gSourceBuffer;
		ioData->mBuffers[0].mDataByteSize = bytesCopied;
	}  
	else
	{
		ioData->mBuffers[0].mData = NULL;      
		ioData->mBuffers[0].mDataByteSize = 0;
		err = noErr;
	}
	playedSize = playedSize + bytesCopied;
	if(playedSize > totalSize) playedSize = totalSize;
	sec = playedSize / 44100 / 4;
		
	if(sec != currentTime) {
		percent = (int)(100.0*playedSize/totalSize);
		printf("\r|");
		for(i=0;i<20;i++) {
			if(percent/5 > i)
				printf("=");
			else if(percent/5 == i)
				printf(">");
			else printf("-");
		}
		printf("| %d:%02d/%d:%02d (%d%%)",sec/60, sec-60*(sec/60), totalTime/60, totalTime-60*(totalTime/60), percent);
		fflush(stdout);
		currentTime = sec;
	}
	
	//if(!buffer2Rest && lastBuffer && !bufferIsFull) playDone = 1;
	
	return err;
}

void *receiveKeyInput(void* param)
{
	int c;
	while(!playDone) {
		c = getchar_unlocked();
		if(c >= '0' && c <= '9') {
			if(muteChannel[c-0x30]) muteChannel[c-0x30] = 0;
			else muteChannel[c-0x30] = 1;
		}
		else if(c >= 'a' && c <= 'n') {
			if(muteChannel[c-0x57]) muteChannel[c-0x57] = 0;
			else muteChannel[c-0x57] = 1;
		}
	}
	return NULL;
}

void setDefaultMuteChannel(int mask)
{
	int i,j;
	for(i=1,j=0;j<24;j++,i*=2) {
		if((mask & i) == 0) muteChannel[j] = 1;
	}
}